#survey_reformatter.py
'''
Takes xlsx file containing additional surveillance data provided by Yoseph and creates entries formatted for Early Warning System clustering.R script.

verbose (False) and verbose_warning (True) are boolean options set in script.

Checks for consistency (fairly thorough but can be improved), and reformats.

Some consistency checks can modify values.

Some expectations can be configured by modifying variable names starting 'expected_'.

Three arguments must be provided:
- region name 'Ethiopia' or 'SouthAsia'.
- input xlsx filename
- output csv filename. Can be a new file, or existing file to append to.
'''

print('Check this is running in python3')

import pandas as pd
import numpy as np
import sys

region = sys.argv[1]

excelDataFilename = sys.argv[2]

outputFileName = sys.argv[3]

#excelDataFilename = '20190902_ETH_YosephA.xlsx'
#excelDataFilename = '20190905_ETH_YosephA.xlsx'
#outputFileName = 'test.csv'
#outputFileName = 'SurveyDataExtraToAddcopy.csv'

# If you want to check all of the comparisons being tested
# (e.g. if checking a new addition is now being used)
# set verbose to True
verbose=False
verbose_warning=True

if verbose:
    def verboseprint(message):
        print(message)
else:
    def verboseprint(message):
        pass

# for reporting errors
def raise_error(message):
    error_str = 'Error raised during'
    print(error_str + message)
    quit()

warning_counter = 0

def raise_warning(message,counting=True):

    global warning_counter
    if counting: warning_counter += 1
    
    warning_str = 'Warning {:d} raised: '.format(warning_counter)
    if verbose_warning:
        print(warning_str + message)

if not len(sys.argv) == 4:
    message = "Should supply exactly three arguments: region name, target excel data file name and output file name. Program called as: "+str(sys.argv)
    raise_error(message)

# define the formats of columns
expected_num_cols = 53

expected_col_formats = {
        'Location ID':int ,
        'HostGenusID':int ,
        'HostGenusNameScientific':str ,
        'HostSpeciesID':int ,
        'HostSpeciesNameScientific':str ,
        'HostCultivarName':str ,
        'PathogenGenusID':int ,
        'PathogenGenusNameScientific':str ,
        'PathogenSpeciesID':int ,
        'PathogenSpeciesNameScientific':str ,
        'UserInit':str ,
        'ObsYear':int ,
        'CountryID':str ,
        'CountryName':str ,
        'RegionID':int ,
        'Latitude':float ,
        'Longitude':float ,
        'Altitude':float ,
        'LocationName':str ,
        'FieldArea':float ,
        'SurveySiteID':int ,
        'SurveySiteName':str ,
        'Observation ID':int ,
        #'ObsDate':int , # this was int in one case, and already datetime in other
        'SurveyorName':str ,
        'InstitutionName':str ,
        'GrowthStageID':int ,
        'GrowthStageName':str ,
        'Comment':str ,
        'Disease ID':int ,
        'PathogenSpeciesName':str ,
        'Severity':int ,
        'SeverityName':str ,
        'Incidence':int ,
        'IncidenceName':str ,
        'InfectionType':int ,
        'InfectionTypeName':str ,
        'Disease ID.1':int ,
        'PathogenSpeciesName.1':str ,
        'Severity.1':int ,
        'SeverityName.1':str ,
        'Incidence.1':int ,
        'IncidenceName.1':str ,
        'InfectionType.1':int ,
        'InfectionTypeName.1':str ,
        'Disease ID.2':int ,
        'PathogenSpeciesName.2':str ,
        'Severity.2':int ,
        'SeverityName.2':str ,
        'Incidence.2':int ,
        'IncidenceName.2':str ,
        'InfectionType.2':int ,
        'InfectionTypeName.2':str }

expected_cols = expected_col_formats.keys()

# open the file
df = pd.read_excel(excelDataFilename, converters = expected_col_formats, na_filter=False, parse_dates=['ObsDate'])

# assume there are only two dimensions
num_rows, num_cols = df.shape

raise_warning('Septoria is not expected from input file')


# functions for consistenty check


def consistency_checks_overall(df):

    # check number of columns
    message = 'checking number of columns equals {:d}'.format(expected_num_cols)
    verboseprint(message)
    try:
        assert num_cols == expected_num_cols

    except AssertionError as error_type:
        
        raise_error(message)

    # when there is more than one survey at the same Location ID, 
    # it has a different observation ID
    message = 'Keeping only last Observation ID at each Location ID (assuming this is latest survey)'
    print(message)
    
    df.drop_duplicates(['Location ID'], keep='last',inplace=True)
    # update number of rows
    num_rows = df.shape[0]

    message = 'checking each row is a unique entry (assuming Location ID tells us this)'
    verboseprint(message)
    try:
        ids = df.loc[:,'Location ID']
        num_ids = ids.unique().size
        assert num_ids == num_rows
    except AssertionError as error_type:
        raise_error(message)

    # TODO: check all other rows for any duplicates

    message = 'check required column names are provided\n'
    actual_cols = df.columns
    message += 'columns are {:s}\n'.format(str(actual_cols))
    message += 'expected columns are {:s}'.format(str(expected_cols))
    verboseprint(message)
    for coln in expected_cols:
        try:
            assert  coln in df.columns
        except AssertionError as error_type:
            message += '\n'+coln+' is expected but not provided'
            raise_error(message)

    print('overall consistency checks complete')
    return

def valid_entries(row,coln,valid_values,message=None):
    '''Given a row of pandas data, and column name, checks that column value 
    is in list of valid values.'''

    if message is None: message = ''

    message += 'for {:s}, accepted range of values is {:s}, '.format(coln, str(valid_values))
    message += 'it actually returns {:s}'.format(str(row[coln]))
    
    verboseprint(message)

    try:

        assert row[coln] in valid_values
    
    except AssertionError as error_type:
        
        raise_error(message)

def valid_entry_span(row,coln,valid_span,alt_value=None,message=None):
    '''Given a row of pandas data, and column name, checks that column value 
    is within a valid span values.'''

    if message is None: message = ''
        
    message += 'for {:s}, accepted range of values is {:s}, '.format(coln,str(valid_span))
    message += 'it actually returns {:s}'.format(str(row[coln]))
    verboseprint(message)

    #if coln=='FieldArea' and row[coln]==-999.99: raise Exception

    try:
        # check that it is not a blank string i.e. ''
        within_span = isinstance(row[coln],float)

        # if a number, check it is within given range
        if within_span:
            within_span = valid_span[0] <= row[coln] <= valid_span[1]
        
        if alt_value is None:
            assert within_span
        elif not within_span:

            message += '\nreplacing with default {:f}'.format(alt_value)
            raise_warning(message,counting=True)
            row[coln] = alt_value

    except AssertionError as error_type:

        raise_error(message)

def consistent_maps(row,coln1,coln2,case1_to_2,message=None):
    '''Checks that the contents of two columns follow expected mapping.'''

    # this case could be replaced with a dict mapping outcome of column 1 to expected outcome of column 2
    case1 = row[coln1]
    case2 = row[coln2]
    
    if case1 in case1_to_2:
        expect_case2 = case1_to_2[case1]
    else:
        # not expecting consistent behaviour so skip
        return

    if message is None: message = ''
    message += '{:s} returns {:s}, so expected {:s} to return {:s}, '.format(coln1,str(case1),coln2,str(expect_case2))
    message += 'it actually returns {:s}'.format(str(case2))
    
    #verboseprint(message)

    try:

        assert case2 == expect_case2

    except AssertionError as error_type:
        
        #raise_error(message)
        raise_warning(message,counting=True)

    return

def consistent_maps_with_replacement(row,coln1,coln2,case1_to_2,remap_1_2,message=None):
    '''Checks that the contents of two columns follow expected mapping.
    If there is a mis-match, the first column name value is replaced.
    case1_to_2 is expected mappings between 1 and 2.
    When they do not map, remap_1_2 tells us how to modify coln1 value
    according to coln2.
    e.g. in this case, the N/A or zero case is replaced with medium
    case1_to_2 = {}
    case1_to_2[-9]=-9
    case1_to_2[0]=0
    case1_to_2[1]=[1,2,3]
    case1_to_2[2]=[1,2,3]
    case1_to_2[3]=[1,2,3]

    from collections import defaultdict
    remap_1_2 = defaultdict(dict)
    #new_y_value = dict[x][y]
    remap_1_2[0][-9] = -9
    remap_1_2[1][-9] = 2
    remap_1_2[2][-9] = 2
    remap_1_2[3][-9] = 2
    remap_1_2[-9][0] = -9
    remap_1_2[1][0] = 2
    remap_1_2[2][0] = 2
    remap_1_2[3][0] = 2
    '''

    # this case could be replaced with a dict mapping outcome of column 1 to expected outcome of column 2
    case1 = row[coln1]
    case2 = row[coln2]
    
    if message is None: message = ''
    message += '{:s} returns {:s}, and {:s} returns {:s}, '.format(coln1,str(case1),coln2,str(case2))
    verboseprint(message)

    if case1 in case1_to_2:
        expect_cases2 = case1_to_2[case1]
    else:
        verboseprint('not expecting consistent behaviour so skip this check')
        return row

    verboseprint('but expected to be in {:s}'.format(str(expect_cases2)))

    if case2 in expect_cases2:
        verboseprint('expected behaviour so return')
        return row

    newval = remap_1_2[case1][case2]

    if case2 in remap_1_2 and case2 == newval:
        verboseprint('unexpected behaviour, case2 already meets new value, no replacement')
        pass

    elif case2 in remap_1_2 and case2 != newval:
        verboseprint('unexpected behaviour, remap case2 based on both cases')

        message += '\nreplacing {:s} with {:s}'.format(coln2,str(newval))
        
        row[coln2] = newval

        raise_warning(message,counting=True)
    
    else:
        message += '\nno solution found yet'
        raise_error(message)

    return row

# defining data structure for consistenty check

# TODO: 
# # higher check: previous data file matches
# perhaps only ever expect additions to older datafiles
# but they might remove entries too.
# Add the following to expected_maps and expected_col_values:
# -check host genum in list.
# -check host species name in list (maybe)
# -check pathogen name
# later, check lat and lon are within sources shapefile
# somehow, check units of field area
# host cultivarname
# check names for special characters: HostSpeciesName, LocationName,HostGenumNameScientific,HostCultivarName
# determine each disease

# defining expected value maps between columns
# This is a list of 3-value tuples: column name1, column name2, and dict that maps between them
expected_maps = []
# for cases where it is appropriate to make a modification for consistency
expected_maps_with_replacement = []

# survey site number and name
expected_siteid_to_sitename = {}
expected_siteid_to_sitename[1]='Farmer field'
expected_siteid_to_sitename[2]='Weed'
expected_siteid_to_sitename[3]='Road side'
expected_siteid_to_sitename[4]='Trial'
expected_siteid_to_sitename[-9]='N/A'
expected_maps += [('SurveySiteID','SurveySiteName',expected_siteid_to_sitename)]
# some other values are probably ok, but not encountered yet

# growth stage id and name
expected_growthid_to_growthname = {}
expected_growthid_to_growthname[1]='Tillering'
expected_growthid_to_growthname[2]='Boot'
expected_growthid_to_growthname[7]='Heading' # note ID is not in growth order
expected_growthid_to_growthname[3]='Flowering'
expected_growthid_to_growthname[4]='Milk'
expected_growthid_to_growthname[5]='Dough'
expected_growthid_to_growthname[6]='Maturity'
expected_growthid_to_growthname[-9]='N/A'
expected_maps += [('GrowthStageID','GrowthStageName',expected_growthid_to_growthname)]

# severity match severity name
expected_severity_to_severityname = {}
expected_severity_to_severityname[-9]='N/A'
expected_severity_to_severityname[0]='None (0)'
expected_severity_to_severityname[1]='Low (less than 20 %)'
expected_severity_to_severityname[2]='Moderate (20 - 40 %)'
expected_severity_to_severityname[3]='High (more than 40 %)'
expected_maps += [('Severity','SeverityName',expected_severity_to_severityname)]

# incidence match incidence name
expected_incidence_to_incidencename = {}
expected_incidence_to_incidencename[-9]='N/A'
expected_incidence_to_incidencename[0]='None (0)'
expected_incidence_to_incidencename[1]='Low (less than 20 %)'
expected_incidence_to_incidencename[2]='Moderate (20 - 40 %)'
expected_incidence_to_incidencename[3]='High (more than 40 %)'
expected_maps += [('Incidence','IncidenceName',expected_incidence_to_incidencename)]

# severity and incidence match when None or N/A.
# when only one is None or N/A, change the other vale to be consistent 
expected_severity_to_incidence = {}
expected_severity_to_incidence[-9]=[-9]
expected_severity_to_incidence[0]=[0]
expected_severity_to_incidence[1]=[1,2,3]
expected_severity_to_incidence[2]=[1,2,3]
expected_severity_to_incidence[3]=[1,2,3]

# create 2D dictionary of acceptable replacement values
from collections import defaultdict
remap_severity_from_incidence = defaultdict(dict)
#new_y_value = dict[x][y]
remap_severity_from_incidence[-9][0] = -9
remap_severity_from_incidence[-9][1] = 1
remap_severity_from_incidence[-9][2] = 2
remap_severity_from_incidence[-9][3] = 3
remap_severity_from_incidence[0][-9] = -9
remap_severity_from_incidence[0][1] = 1
remap_severity_from_incidence[0][2] = 2
remap_severity_from_incidence[0][3] = 3
remap_severity_from_incidence[1][-9] = 2
remap_severity_from_incidence[1][0] = 2
remap_severity_from_incidence[2][-9] = 2
remap_severity_from_incidence[2][0] = 2
remap_severity_from_incidence[3][-9] = 2
remap_severity_from_incidence[3][0] = 2

expected_maps_with_replacement += [('Severity','Incidence',expected_severity_to_incidence, remap_severity_from_incidence)]
#expected_maps_with_replacement += [('Severity.1','Incidence.1',expected_severity_to_incidence)] # leaf rust not important
expected_maps_with_replacement += [('Severity.2','Incidence.2',expected_severity_to_incidence,remap_severity_from_incidence)]

expected_incidence_to_severity = expected_severity_to_incidence

remap_incidence_from_severity = remap_severity_from_incidence

expected_maps_with_replacement += [('Incidence','Severity',expected_incidence_to_severity, remap_incidence_from_severity)]
#expected_maps_with_replacement += [('Incidence.1','Severity.1',expected_incidence_to_severity)]  # leaf rust not important
expected_maps_with_replacement += [('Incidence.2','Severity.2',expected_incidence_to_severity, remap_incidence_from_severity)]

# do the same with names of severity and incidence

# use a simpler representation of the names
n = {}
n[0] = 'None (0)'
n[1] = 'Low (less than 20 %)'
n[2] = 'Moderate (20 - 40 %)'
n[3] = 'High (more than 40 %)'
n[-9] = 'N/A'

expected_severityname_to_incidencename = {}
expected_severityname_to_incidencename[n[-9]]=[n[-9]]
expected_severityname_to_incidencename[n[0]]=[n[0]]
expected_severityname_to_incidencename[n[1]]=[n[1],n[2],n[3]]
expected_severityname_to_incidencename[n[2]]=[n[1],n[2],n[3]]
expected_severityname_to_incidencename[n[3]]=[n[1],n[2],n[3]]

remap_severityname_from_incidencename = defaultdict(dict)
#new_y_value = dict[x][y]
remap_severityname_from_incidencename[n[0]][n[-9]] = n[-9]
remap_severityname_from_incidencename[n[0]][n[1]] = n[1]
remap_severityname_from_incidencename[n[0]][n[2]] = n[2]
remap_severityname_from_incidencename[n[0]][n[3]] = n[3]
remap_severityname_from_incidencename[n[1]][n[-9]] = n[2]
remap_severityname_from_incidencename[n[2]][n[-9]] = n[2]
remap_severityname_from_incidencename[n[3]][n[-9]] = n[2]
remap_severityname_from_incidencename[n[-9]][n[0]] = n[-9]
remap_severityname_from_incidencename[n[-9]][n[1]] = n[1]
remap_severityname_from_incidencename[n[-9]][n[2]] = n[2]
remap_severityname_from_incidencename[n[-9]][n[3]] = n[3]
remap_severityname_from_incidencename[n[1]][n[0]] = n[2]
remap_severityname_from_incidencename[n[2]][n[0]] = n[2]
remap_severityname_from_incidencename[n[3]][n[0]] = n[2]

expected_maps_with_replacement += [('SeverityName','IncidenceName',expected_severityname_to_incidencename, remap_severityname_from_incidencename)]
#expected_maps_with_replacement += [('SeverityName.1','IncidenceName.1',expected_severityname_to_incidencename)] # leaf rust not important
expected_maps_with_replacement += [('SeverityName.2','IncidenceName.2',expected_severityname_to_incidencename,remap_severityname_from_incidencename)]

expected_incidencename_to_severityname = expected_severityname_to_incidencename

remap_incidencename_from_severityname = remap_severityname_from_incidencename

expected_maps_with_replacement += [('IncidenceName','SeverityName',expected_incidencename_to_severityname, remap_incidencename_from_severityname)]
#expected_maps_with_replacement += [('IncidenceName.1','SeverityName.1',expected_incidencename_to_severityname)]  # leaf rust not important
expected_maps_with_replacement += [('IncidenceName.2','SeverityName.2',expected_incidencename_to_severityname, remap_incidencename_from_severityname)]


# infection stage id and name
expected_infectionid_to_infectionname = {}
expected_infectionid_to_infectionname[1]='R'
expected_infectionid_to_infectionname[2]='R-MR'
expected_infectionid_to_infectionname[3]='MR'
expected_infectionid_to_infectionname[4]='MR-M'
expected_infectionid_to_infectionname[5]='M (MR-MS)'
expected_infectionid_to_infectionname[6]='M-MS'
expected_infectionid_to_infectionname[7]='MS'
expected_infectionid_to_infectionname[8]='MS-S'
expected_infectionid_to_infectionname[9]='S'
expected_infectionid_to_infectionname[-9]='N/A'
expected_maps += [('InfectionType','InfectionTypeName',expected_infectionid_to_infectionname)]

# disease id match pathogen species name
expected_diseaseid_to_pathogenname = {}
expected_diseaseid_to_pathogenname[1]='Stem Rust'
expected_diseaseid_to_pathogenname[2]='Leaf Rust'
expected_diseaseid_to_pathogenname[3]='Yellow Rust'
expected_maps += [('Disease ID','PathogenSpeciesName',expected_diseaseid_to_pathogenname)]

message = 'expected maps are:\n'
message += str(expected_maps).replace('),','),\n')
verboseprint(message)

# defining expected values in column
# using definitions of map dicts where they are complete

expected_col_values = {}
expected_col_values['SurveySiteID'] = expected_siteid_to_sitename.keys() 
expected_col_values['SurveySiteName'] = expected_siteid_to_sitename.values() # some other values are probably ok, but not encountered yet
expected_col_values['GrowthStageID'] = expected_growthid_to_growthname.keys()
expected_col_values['GrowthStageName'] = expected_growthid_to_growthname.values()

print('Expect information on only three diseases: stem, leaf and yellow rusts')
expected_col_values['Disease ID'] = [1]
expected_col_values['Disease ID.1'] = [2]
expected_col_values['Disease ID.2'] = [3]

expected_col_values['Severity'] = expected_severity_to_severityname.keys()
expected_col_values['Incidence'] = expected_incidence_to_incidencename.keys()
expected_col_values['SeverityName'] = expected_severity_to_severityname.values()
expected_col_values['IncidenceName'] = expected_incidence_to_incidencename.values()
expected_col_values['InfectionType'] = expected_infectionid_to_infectionname.keys()
expected_col_values['InfectionTypeName'] = expected_infectionid_to_infectionname.values()

expected_col_values['Severity.1'] = expected_col_values['Severity']
expected_col_values['Incidence.1'] = expected_col_values['Incidence']
expected_col_values['SeverityName.1'] = expected_col_values['SeverityName']
expected_col_values['IncidenceName.1'] = expected_col_values['IncidenceName']
expected_col_values['InfectionType.1'] = expected_col_values['InfectionType']
expected_col_values['InfectionTypeName.1'] = expected_col_values['InfectionTypeName']

expected_col_values['Severity.2'] = expected_col_values['Severity']
expected_col_values['Incidence.2'] = expected_col_values['Incidence']
expected_col_values['SeverityName.2'] = expected_col_values['SeverityName']
expected_col_values['IncidenceName.2'] = expected_col_values['IncidenceName']
expected_col_values['InfectionType.2'] = expected_col_values['InfectionType']
expected_col_values['InfectionTypeName.2'] = expected_col_values['InfectionTypeName']

expected_col_values['PathogenSpeciesName'] = expected_diseaseid_to_pathogenname.values()

expected_col_values['ObsYear'] = range(2011,2019+1)

# defining expected limits of values in certain columns
expected_col_value_limits = {}
expected_col_value_limits['FieldArea'] = [0.001,100]
expected_col_value_limits['Altitude'] = [-1000,5000]

# defining values to be applied if outside of expected limits.
expected_col_value_alternative = {}
expected_col_value_alternative['FieldArea'] = 0.01

if region == 'Ethiopia':
    expected_col_values['CountryName'] = ['Ethiopia']

    #expected_col_value_limits['Longitude'] = [32.,48.1] # limits of the country of Ethiopia
    #expected_col_value_limits['Latitude'] = [3.,15.] # limits of the country of Ethiopia
    expected_col_value_limits['Longitude'] = [34.9,43.] # limits of the fixed shapefile of sources for Ethiopia
    expected_col_value_limits['Latitude'] = [4.7,15.] # limits of the fixed shapefile of sources for Ethiopia

elif region == 'SouthAsia':
    expected_col_values['CountryName'] = ['Bangladesh','Bhutan','India','Nepal','Pakistan']

    expected_col_value_limits['Longitude'] = [65.,98.] # limits of the fixed shapefile of sources for South Asia
    expected_col_value_limits['Latitude'] = [7.,36.] # limits of the fixed shapefile of sources for South Asia


# run consistency checks 


consistency_checks_overall(df)

def consistent_fields_per_row(row):
    '''Checks where there is expectation of entries in one column mapping to entries in other.'''

    row_id = row['Location ID']
    message = 'For row with Location ID {:d}:\n'.format(row_id)
    
    # check entries in defined columns are as expected
    for coln,valid_values in expected_col_values.items():
        valid_entries(row,coln,valid_values,message=message)

    # check entries in defined columns are within expected limits
    for coln,valid_span in expected_col_value_limits.items():
        valid_entry_span(row,coln,valid_span,alt_value=expected_col_value_alternative.get(coln,None),message=message)

    # check values in column correspond to value occurring in another
    for coln1,coln2,namemap in expected_maps:
        consistent_maps(row,coln1,coln2,namemap,message=message)

    # some other column comparisons, but mis-matches require modification
    for coln1,coln2,namemap,remap in expected_maps_with_replacement:
        row = consistent_maps_with_replacement(row,coln1,coln2,namemap,remap,message=message)
    
    return row

print('Checking consistent entries per row')

df2 = df.apply(consistent_fields_per_row,axis=1)

df = df2


# define functions to convert to new dataframe, matching example csv.


def convert_ObsDate_to_datestr(series):

    if series.dtype == 'int64':

        date = pd.TimedeltaIndex(series, unit='d') + pd.datetime(1899, 12, 30)

        datestr = date.strftime('%d-%b-%Y')
    
    elif isinstance(series[0],pd.datetime):
        
        # excel date format was read successfully
        # no need to convert from integer

        datestr = series.dt.strftime('%d-%b-%Y')

        pass


    return datestr

def convert_to_small_case(series):

    return series.str.lower()

def convert_severity(series):
    
    # the output values are chosen to be within the range described by input
    # The supplement to Sader et al. paper suggests the calculation only cares
    # about which range it is in, not the exact number.
    # So, the EWS calculation appears to want a number that is too specific, 
    # when 'low', 'moderate' or 'high' would be sufficient.
    map_severity_in_out = {
            'Low (less than 20 %)':'10',
            'Moderate (20 - 40 %)':'30',
            'High (more than 40 %)':'50',
            'None (0)':'0',
            'N/A':'na'}

    return series.map(map_severity_in_out)

def convert_incidence(series):

    map_severity_in_out = {
            'Low (less than 20 %)':'low',
            'Moderate (20 - 40 %)':'medium',
            'High (more than 40 %)':'high',
            'None (0)':'none',
            'N/A':'na'}

    return series.map(map_severity_in_out)

def convert_reaction(series):

    # lower case
    series_out = series.str.lower()

    series_out = series_out.apply(lambda cell: cell.replace('/',''))

    # any action is quite arbitrary, not well founded
    # Sader et al supplement indicates this is not needed for source strength calcs
    
    # so, keep but do not modify
    
    return series_out

def do_nothing(series):
    return series


# defining data structure to convert to new dataframe


# the required order of columns in output
default_output_cols = [
        'SubmissionDate',
        'start',
        'end',
        'today',
        'deviceid',
        'subscriberid',
        'imei',
        'phonenumber',
        'username',
        'surveyor_infromation-country',
        'surveyor_infromation-surveyor_name',
        'surveyor_infromation-institution',
        'survey_infromation-location_name',
        'survey_infromation-location-Latitude',
        'survey_infromation-location-Longitude',
        'survey_infromation-location-Altitude',
        'survey_infromation-location-Accuracy',
        'survey_infromation-survey_date',
        'site_information-survey_site',
        'site_information-crop',
        'site_information-field_area',
        'site_information-variety',
        'site_information-growth_stage',
        'other_crop',
        'stem_rust-stemrust_incidence',
        'stem_rust-Stemrust_severity',
        'stem_rust-stemrust_host_plant_reaction',
        'leaf_rust-leafrust_incidence',
        'leaf_rust-leafrust_severity',
        'leaf_rust-leafrust_host_plant_reaction',
        'yellow_rust-yellowrust_incidence',
        'yellow_rust-yellowrust_severity',
        'yellow_rust-yellowrust_host_plant_reaction',
        'septoria-septoria_incidence',
        'septoria-septoria_severity',
        'other_diseases_group-other_diseases',
        'score_diseases_count',
        'SET-OF-score_diseases',
        'samples_collected',
        'samples_type',
        'sample_size-number_stemrust_live',
        'sample_size-number_stemrust_dead_dna',
        'sample_size-number_yellowrust_live',
        'sample_size-number_yellowrust_dead',
        'sample_size-number_leafrust_live',
        'sample_size-using_barcode',
        'live_stemrust_samples_count',
        'SET-OF-live_stemrust_samples',
        'dead_stemrust_samples_count',
        'SET-OF-dead_stemrust_samples',
        'live_yellowrust_samples_count',
        'SET-OF-live_yellowrust_samples',
        'dead_yellowrust_samples_count',
        'SET-OF-dead_yellowrust_samples',
        'live_leafrust_samples_count',
        'SET-OF-live_leafrust_samples',
        'comment',
        'meta-instanceID',
        'KEY']

default_output_col_values = {
        'SubmissionDate':'?',
        'start':'?',
        'end':'?',
        'today':'?',
        'deviceid':999,
        'subscriberid':999,
        'imei':999,
        'phonenumber':999,
        'username':999,
        #'surveyor_infromation-country':'?', # commented lines are expected from input spreadsheet
        #'surveyor_infromation-surveyor_name':999,
        #'surveyor_infromation-institution':999,
        #'survey_infromation-location_name':999,
        #'survey_infromation-location-Latitude':'?',
        #'survey_infromation-location-Longitude':'?',
        #'survey_infromation-location-Altitude':'?',
        'survey_infromation-location-Accuracy':-999,
        'survey_infromation-survey_date':'?',
        #'site_information-survey_site':'?',
        'site_information-crop':'NA',
        #'site_information-field_area':'?',
        #'site_information-variety':'?',
        #'site_information-growth_stage':'?',
        'other_crop':'-999',
        #'stem_rust-stemrust_incidence':'?',
        #'stem_rust-Stemrust_severity':'?',
        #'stem_rust-stemrust_host_plant_reaction':'?',
        #'leaf_rust-leafrust_incidence':'?',
        #'leaf_rust-leafrust_severity':'?',
        #'leaf_rust-leafrust_host_plant_reaction':'?',
        #'yellow_rust-yellowrust_incidence':'?',
        #'yellow_rust-yellowrust_severity':'?',
        #'yellow_rust-yellowrust_host_plant_reaction':'?',
        'septoria-septoria_incidence':'na',
        'septoria-septoria_severity':'na',
        'other_diseases_group-other_diseases':-999,
        'score_diseases_count':-999,
        'SET-OF-score_diseases':-999,
        'samples_collected':-999,
        'samples_type':-999,
        'sample_size-number_stemrust_live':-999,
        'sample_size-number_stemrust_dead_dna':-999,
        'sample_size-number_yellowrust_live':-999,
        'sample_size-number_yellowrust_dead':-999,
        'sample_size-number_leafrust_live':-999,
        'sample_size-using_barcode':-999,
        'live_stemrust_samples_count':-999,
        'SET-OF-live_stemrust_samples':-999,
        'dead_stemrust_samples_count':-999,
        'SET-OF-dead_stemrust_samples':-999,
        'live_yellowrust_samples_count':-999,
        'SET-OF-live_yellowrust_samples':-999,
        'dead_yellowrust_samples_count':-999,
        'SET-OF-dead_yellowrust_samples':-999,
        'live_leafrust_samples_count':-999,
        'SET-OF-live_leafrust_samples':-999,
        'comment':-999,
        'meta-instanceID':-999,
        'KEY':-999}

# which output columns can be got from input file
expected_map_output_input = {
        'SubmissionDate':'ObsDate', # this needs a date reformat
        'start':'ObsDate', # this needs a date reformat
        'end':'ObsDate', # this needs a date reformat
        'today':'ObsDate', # this needs a date reformat
        'site_information-variety':'HostCultivarName',
        'surveyor_infromation-country':'CountryName',
        'survey_infromation-location_name':'LocationName',
        'survey_infromation-location-Latitude':'Latitude',
        'survey_infromation-location-Longitude':'Longitude',
        'survey_infromation-location-Altitude':'Altitude',
        'site_information-field_area':'FieldArea',
        'site_information-survey_site':'SurveySiteName',
        'surveyor_infromation-surveyor_name':'SurveyorName',
        'surveyor_infromation-institution':'InstitutionName',
        'survey_infromation-survey_date':'ObsDate', # this needs a date reformat
        'site_information-growth_stage':'GrowthStageName', # this needs to be put as small caps
        'stem_rust-Stemrust_severity':'SeverityName', # this needs reformatting to appropriate values
        'stem_rust-stemrust_incidence':'IncidenceName',  # this needs reformatting to appropriate strings
        'stem_rust-stemrust_host_plant_reaction':'InfectionTypeName', # this needs reformatting to appropriate strings
        'leaf_rust-leafrust_severity':'SeverityName.1', # this needs reformatting to appropriate values
        'leaf_rust-leafrust_incidence':'IncidenceName.1',  # this needs reformatting to appropriate strings
        'leaf_rust-leafrust_host_plant_reaction':'InfectionTypeName.1',  # this needs reformatting to appropriate strings
        'yellow_rust-yellowrust_severity':'SeverityName.2', # this needs reformatting to appropriate values
        'yellow_rust-yellowrust_incidence':'IncidenceName.2',  # this needs reformatting to appropriate strings
        'yellow_rust-yellowrust_host_plant_reaction':'InfectionTypeName.2' } # this needs reformatting to appropriate strings

# what conversion is needed from input column to output column
expected_map_output_input_function = {
        'SubmissionDate':convert_ObsDate_to_datestr, 
        'start':convert_ObsDate_to_datestr,
        'end':convert_ObsDate_to_datestr, 
        'today':convert_ObsDate_to_datestr, 
        'site_information-variety':do_nothing,
        'surveyor_infromation-country':do_nothing,
        'survey_infromation-location-Latitude':do_nothing,
        'survey_infromation-location-Longitude':do_nothing,
        'survey_infromation-location-Altitude':do_nothing, 
        'survey_infromation-location_name':do_nothing,
        'site_information-field_area':do_nothing,  # TODO: replace NaN with fill value
        'site_information-survey_site':do_nothing,
        'surveyor_infromation-surveyor_name':do_nothing,
        'surveyor_infromation-institution':do_nothing,
        'survey_infromation-survey_date':convert_ObsDate_to_datestr,
        'site_information-growth_stage':convert_to_small_case,
        'stem_rust-Stemrust_severity':convert_severity, 
        'stem_rust-stemrust_incidence':convert_incidence, 
        'stem_rust-stemrust_host_plant_reaction':convert_reaction, # this needs reformatting to appropriate strings
        'leaf_rust-leafrust_severity':convert_severity, 
        'leaf_rust-leafrust_incidence':convert_incidence,
        'leaf_rust-leafrust_host_plant_reaction':convert_reaction,  # this needs reformatting to appropriate strings
        'yellow_rust-yellowrust_severity':convert_severity,
        'yellow_rust-yellowrust_incidence':convert_incidence, 
        'yellow_rust-yellowrust_host_plant_reaction':convert_reaction } # this needs reformatting to appropriate strings

expected_col_formats = {
        'SubmissionDate':str,
        'start':str,
        'end':str,
        'today':str,
        'deviceid':str,
        'subscriberid':str,
        'imei':str,
        'phonenumber':str,
        'username':str,
        'surveyor_infromation-country':str,
        'surveyor_infromation-surveyor_name':str,
        'surveyor_infromation-institution':str,
        'survey_infromation-location_name':str,
        'survey_infromation-location-Latitude':str,
        'survey_infromation-location-Longitude':str,
        'survey_infromation-location-Altitude':str,
        'survey_infromation-location-Accuracy':str,
        'survey_infromation-survey_date':str,
        'site_information-survey_site':str,
        'site_information-crop':str,
        'site_information-field_area':str,
        'site_information-variety':str,
        'site_information-growth_stage':str,
        'other_crop':str,
        'stem_rust-stemrust_incidence':str,
        'stem_rust-Stemrust_severity':str,
        'stem_rust-stemrust_host_plant_reaction':str,
        'leaf_rust-leafrust_incidence':str,
        'leaf_rust-leafrust_severity':str,
        'leaf_rust-leafrust_host_plant_reaction':str,
        'yellow_rust-yellowrust_incidence':str,
        'yellow_rust-yellowrust_severity':str,
        'yellow_rust-yellowrust_host_plant_reaction':str,
        'septoria-septoria_incidence':str,
        'septoria-septoria_severity':str,
        'other_diseases_group-other_diseases':str,
        'score_diseases_count':str,
        'SET-OF-score_diseases':str,
        'samples_collected':str,
        'samples_type':str,
        'sample_size-number_stemrust_live':str,
        'sample_size-number_stemrust_dead_dna':str,
        'sample_size-number_yellowrust_live':str,
        'sample_size-number_yellowrust_dead':str,
        'sample_size-number_leafrust_live':str,
        'sample_size-using_barcode':str,
        'live_stemrust_samples_count':str,
        'SET-OF-live_stemrust_samples':str,
        'dead_stemrust_samples_count':str,
        'SET-OF-dead_stemrust_samples':str,
        'live_yellowrust_samples_count':str,
        'SET-OF-live_yellowrust_samples':str,
        'dead_yellowrust_samples_count':str,
        'SET-OF-dead_yellowrust_samples':str,
        'live_leafrust_samples_count':str,
        'SET-OF-live_leafrust_samples':str,
        'comment':str,
        'meta-instanceID':str,
        'KEY':str}


# calculate output dataframe

#TODO:
# Handle column formats
# check date (once calculated) are valid. not future. (this is low priority because the year is already checked with ObsYear)

dfout = pd.DataFrame()

# add required columns, one at a time, in order
for outcoln in default_output_cols:
    
    if outcoln in expected_map_output_input:
        # take date from input, convert where necessary

        assert outcoln in expected_map_output_input_function

        incoln = expected_map_output_input[outcoln]

        func = expected_map_output_input_function[outcoln]

        dfout[outcoln] = func(df[incoln])
    
    else:

        # default values must be provided
        default_value = default_output_col_values[outcoln]
        dfout[outcoln] = pd.Series([default_value]*num_rows)

dfout = dfout.astype(expected_col_formats)

# append/write to file
with open(outputFileName, 'a') as f:

    # writing header if file is new/empty
    header = f.tell()==0

    dfout.to_csv(f,header=header,index=False)

print('{:d} warnings'.format(warning_counter))

print('end of script')
